// This software is Copyright (C) 2005 by Andrew Firth, University of Otago, 
// Dunedin, New Zealand. All rights reserved.

// Programming for making running means of columns 2 and 3 in input logfile.
// Command line input: file name and halfwindow size n (window is 2n + 1).
// Leaves column 1 unchanged.  Takes running mean of column 2 (conservation 
//   scores).  Sqrt of sum of squares for column 3 (stddev of conservation 
//   scores).
// Assumes input file is in nucleotide order.
// To skip gaps (etc) just miss those nucleotides out of the input file.

#include <iostream>
#include <cmath>
#include <fstream>
#include <cstdlib>
#include <iomanip>
#include<cstring>
using namespace std;
#define maxlength 50000

int main(int argc, char* argv[])
{
  if (4 != argc) {
    cerr << "Aborting; use '" << argv[0] 
	 << " logfile window_half_size circular'.\n";
    exit(EXIT_FAILURE);
  }

  int verbose = 0;

  int j, k, nlines, col1[maxlength], n, hfw, circular;
  double col2[maxlength], col3[maxlength], sum1, sum2;

  hfw = atoi(argv[2]);
  circular = atoi(argv[3]);

  // Initialize arrays
  for (k = 0; k < maxlength; ++k) {
    col1[k] = 0;
    col2[k] = col3[k] = 0.;
  }

  // Open log file.
  ifstream logfile(argv[1]); 
  if (!logfile) {
    cerr << "Aborting: can't find log file '" << argv[1] << "'.\n";
    exit(EXIT_FAILURE);
  }

  // Number of lines in log file
  nlines = -1;
  while (logfile.ignore(1000, '\n')) {
    ++nlines;
  }
  logfile.clear();
  logfile.seekg(0);
  if (verbose) {
    cout << "There are " << nlines << " data lines.\n\n";
  }

  // Read in logfile
  for (k = 0; k < nlines; ++k) {
    logfile >> col1[k];
    logfile >> col2[k];
    logfile >> col3[k];
    logfile.ignore(1000, '\n');
  }

  // Running mean
  for (k = 0; k < nlines; ++k) {
    sum1 = sum2 = 0.;
    n = 0;
    for (j = k - hfw; j <= k + hfw; ++j) { 
      if (!circular && j >= 0 && j < nlines || circular) {
	sum1 += col2[(j+nlines)%nlines];
	sum2 += col3[(j+nlines)%nlines] * col3[(j+nlines)%nlines];
	n += 1;
      }
    }
    cout << col1[k] << " " << sum1/float(n) << " " << sqrt(sum2)/float(n) 
	 << "\n";
  }

}
